#include "remote_headset.hpp"
#include "vr_application.hpp"
#include <imgui.h>

#include <glm/gtx/string_cast.hpp>

RemoteHeadset::RemoteHeadset()
{
    this->controller_states.fill(CONTROLLER_STATE_DETACHED);
    this->transport_controller_states.fill(CONTROLLER_STATE_DETACHED);

    for (std::array<PacketButtonState, BUTTON_ID_MAX_COUNT>& buttons : this->controller_buttons)
    {
        buttons.fill(BUTTON_STATE_RELEASED);   
    }
}

bool RemoteHeadset::on_setup_instance(lava::frame_config& config)
{
    // Required in oder to get:
    // VK_KHR_external_fence,
    // VK_KHR_external_fence_capabilities,
    // VK_KHR_get_physical_device_properties2
    config.info.req_api_version = lava::api_version::v1_1;

    return true;
}

bool RemoteHeadset::on_setup_device(lava::device::create_param& parameters)
{
    if (!setup_extern_fence(this->get_application()->get_instance(), parameters.extensions))
    {
        return false;
    }

    if (!setup_encoder(parameters))
    {
        return false;
    }

    return true;
}

bool RemoteHeadset::on_create()
{
    this->statistic_send_bitrate = make_statistic("Send-Bitrate (Mbps)", 128);
    this->statistic_send_queue_size = make_statistic("Send-Queue (KBytes)", 128);
    this->statistic_receive_bitrate = make_statistic("Receive-Bitrate (Mbps)", 128);
    this->statistic_receive_queue_size = make_statistic("Receive-Queue (KBytes)", 128);
    this->statistic_latency = make_latency_statistic("latency");

    StereoStrategy::Ptr strategy = this->get_application()->get_stereo_strategy();
    this->frame_id_count = strategy->get_max_remote_frame_ids();

    if (!convert_strategy(this->get_application()->get_stereo_strategy_type(), this->strategy_id))
    {
        return false;
    }

    if (!this->create_transport())
    {
        lava::log()->error("Remote Headset: Can't create server!");

        return false;
    }

    lava::log()->debug("Waiting for connection");

    if (!this->transport->wait_connect())
    {
        lava::log()->error("Remote Headset: Incomplete setup handshake!");

        return false;    
    }

    this->update_values();

    if (!this->create_framebuffers())
    {
        lava::log()->error("Remote Headset: Can't create framebuffers!");

        return false;        
    }

    if (!this->create_encoders())
    {
        lava::log()->error("Remote Headset: Can't create encoders!");

        return false;
    }

    return true;
}

void RemoteHeadset::on_destroy()
{
    if (this->transport != nullptr)
    {
        this->transport->destroy();
    }

    if (this->statistic_latency != nullptr)
    {
        statistic_latency->write_to_file();
    }

    this->destroy_framebuffers();
    this->destroy_encoders();
}

bool RemoteHeadset::on_interface()
{
    if (this->transport->get_state() != TRANSPORT_STATE_CONNECTED)
    {
        return false;
    }

    ImGui::DragFloat("Movement Speed", &this->movement_speed, 1.0f, 0.0f, 1000.0);

    ImGui::Separator();

    std::vector<const char*> encoder_mode_names;
    encoder_mode_names.push_back("Constant Bitrate");
    encoder_mode_names.push_back("Constant Quality");

    if (ImGui::Combo("Encoder Mode", (int32_t*)&this->encoder_mode, encoder_mode_names.data(), encoder_mode_names.size()))
    {
        for (Encoder::Ptr encoder : this->encoders)
        {
            encoder->set_mode((EncoderMode)this->encoder_mode);
        }
    }

    ImGui::SliderInt("Input-Rate (Fps)", (int32_t*) &this->encoder_input_rate, 1, 180);

    if (ImGui::SliderInt("Key-Rate (Frames)", (int32_t*)&this->encoder_key_rate, 1, 180))
    {
        for (Encoder::Ptr encoder : this->encoders)
        {
            encoder->set_key_rate(this->encoder_key_rate);
        }
    }

    if (this->encoder_mode == ENCODER_MODE_CONSTANT_BITRATE)
    {
        if (ImGui::SliderFloat("Bit-Rate (Mbps)", &this->encoder_bit_rate, 0.1, 10.0))
        {
            for (Encoder::Ptr encoder : this->encoders)
            {
                encoder->set_bit_rate(this->encoder_bit_rate * 1000);
            }
        }

        if (ImGui::SliderInt("Frame-Rate (Fps)", (int32_t*)&this->encoder_frame_rate, 1, 180))
        {
            for (Encoder::Ptr encoder : this->encoders)
            {
                encoder->set_frame_rate(this->encoder_frame_rate);
            }
        }
    }

    else if (this->encoder_mode == ENCODER_MODE_CONSTANT_QUALITY)
    {
        if (ImGui::SliderInt("Quality I-Frame", (int32_t*)&this->encoder_quality_iframe, 1, 51))
        {
            for (Encoder::Ptr encoder : this->encoders)
            {
                encoder->set_quality_iframe(this->encoder_quality_iframe);
            }
        }

        if (ImGui::SliderInt("Quality P-Frame", (int32_t*)&this->encoder_quality_pframe, 1, 51))
        {
            for (Encoder::Ptr encoder : this->encoders)
            {
                encoder->set_quality_pframe(this->encoder_quality_pframe);
            }
        }
    }

    else
    {
        lava::log()->error("Remote Headset: Unknown encoder mode!");

        return false;
    }

    /*if (ImGui::SliderFloat("Send-Rate Limit (Mbps)", &this->send_bit_rate_limit, 1.0, 20.0))
    {
        this->transport->set_max_bitrate_send(this->send_bit_rate_limit);
    }*/

    if (ImGui::SliderInt("Send-Queue Limit (KBytes)", (int32_t*)&this->send_queue_size_limit, 1, 1000))
    {
        this->transport->set_max_send_queue_size(this->send_queue_size_limit * 1000);
    }

    ImGui::Separator();

    this->statistic_send_bitrate->interface();
    this->statistic_receive_bitrate->interface();

    this->statistic_send_queue_size->interface();
    this->statistic_receive_queue_size->interface();

    std::unique_lock<std::mutex> lock(this->transport_mutex);
    this->statistic_latency->interface();
    lock.unlock();

    return true;
}

bool RemoteHeadset::on_update(lava::delta delta_time)
{
    if (this->transport->get_state() != TRANSPORT_STATE_CONNECTED)
    {
        return false;
    }

    this->update_values();
    this->update_bitrates();
    this->update_stage(delta_time);
    
    this->encoder_last_submit += delta_time;
    this->encoder_enable_submit = false;

    if (this->encoder_last_submit > (1.0 / (float)this->encoder_input_rate))
    {
        this->frame_number++;
        this->encoder_enable_submit = true;
        this->encoder_last_submit = 0.0f;
    }

    return true;
}

void RemoteHeadset::submit_frame(VkCommandBuffer command_buffer, VkImageLayout frame_layout, FrameId frame_id)
{
    if (frame_id >= this->frame_id_count)
    {
        lava::log()->error("Remote Headset: Invalid frame id during frame submission!");

        return;
    }

    if (!this->encoder_enable_submit)
    {
        return;        
    }

    lava::renderer& renderer = this->get_application()->get_renderer();
    lava::image::ptr framebuffer = this->framebuffers[frame_id];

    PassTimer::Ptr pass_timer = this->get_application()->get_pass_timer();
    Encoder::Ptr encoder = this->encoders[frame_id];    
    uint32_t frame_number = this->frame_number;
    uint32_t transform_id = this->transform_id;

    pass_timer->begin_pass(command_buffer, "Encode Setup " + std::to_string(frame_id));
    bool result = encoder->encode_frame(command_buffer, renderer, framebuffer, frame_layout, [this, frame_number, frame_id, transform_id](const std::span<uint8_t>& content, EncoderFrameType type)
    {
        switch (type)
        {
        case ENCODER_FRAME_TYPE_CONFIG:
            this->transport->send_frame_config_nal(frame_id, content);
            break;
        case ENCODER_FRAME_TYPE_FRAME:
            this->transport->send_frame_nal(frame_number, frame_id, transform_id, content);
            break;
        case ENCODER_FRAME_TYPE_CONTROL:
        default:
            break;
        }
    });
    pass_timer->end_pass(command_buffer);

    if (!result)
    {
        lava::log()->error("Remote Headset: Error during encode submission!");
    }
}

void RemoteHeadset::submit_metadata(const FrameMetadata& metadata)
{
    this->transport->send_frame_metadata(this->frame_number, metadata);
}

VkFormat RemoteHeadset::get_format() const
{
    return VK_FORMAT_R8G8B8A8_SRGB;
}

const glm::uvec2& RemoteHeadset::get_resolution() const
{
    return this->resolution;
}

const glm::mat4& RemoteHeadset::get_view_matrix() const
{
    return this->view_matrix;
}

const glm::mat4& RemoteHeadset::get_head_to_eye_matrix(Eye eye) const
{
    return this->head_to_eye_matrices[eye];
}

const glm::mat4& RemoteHeadset::get_projection_matrix(Eye eye) const
{
    return this->projection_matrices[eye];
}

const glm::mat4& RemoteHeadset::get_controller_matrix(Controller controller) const
{
    return this->controller_matrices[controller];
}

lava::image::ptr RemoteHeadset::get_framebuffer(FrameId frame_id) const
{
    return this->framebuffers[frame_id];
}

lava::image::ptr RemoteHeadset::get_framebuffer_array() const
{
    return this->framebuffer_array;
}

const char* RemoteHeadset::get_name() const
{
    return "Remote Headset";
}

bool RemoteHeadset::is_attached(Controller controller) const
{
    return this->controller_states[controller] == CONTROLLER_STATE_ATTACHED;
}

bool RemoteHeadset::is_remote() const
{
    return true;
}

bool RemoteHeadset::create_transport()
{
    this->transport = make_transport();

    this->transport->set_on_setup([this](const glm::u32vec2& resolution)
    {
        this->on_setup(resolution);
    });
    this->transport->set_on_projection_change([this](const glm::mat4& left_eye_projection, const glm::mat4& right_eye_projection, const glm::mat4& left_head_to_eye, const glm::mat4& right_head_to_eye)
    {
        this->on_projection_change(left_eye_projection, right_eye_projection, left_head_to_eye, right_head_to_eye);
    });
    this->transport->set_on_head_transform([this](PacketTransformId transform_id, const glm::mat4& head_transform)
    {
        this->on_head_transform(transform_id, head_transform);
    });
    this->transport->set_on_controller_transform([this](PacketControllerId controller_id, const glm::mat4& controller_transform)
    {
        this->on_controller_transform(controller_id, controller_transform);
    });
    this->transport->set_on_controller_event([this](PacketControllerId controller_id, PacketControllerState controller_state)
    {
        this->on_controller_event(controller_id, controller_state);
    });
    this->transport->set_on_controller_button([this](PacketControllerId controller_id, PacketButtonId button_id, PacketButtonState button_state)
    {
        this->on_controller_button(controller_id, button_id, button_state);
    });
    this->transport->set_on_controller_thumbstick([this](PacketControllerId controller_id, const glm::vec2& thumbstick)
    {
        this->on_controller_thumbstick(controller_id, thumbstick);
    });
    this->transport->set_on_latency([this](PacketFrameNumber frame_number, PacketFrameId frame_id, PacketTransformId transform_id, double timestamp_transfrom_query, double timestamp_server_response, double timestamp_frame_decoded, double timestamp_command_construct, double timestamp_command_executed)
    {
        this->on_latency(frame_number, frame_id, transform_id, timestamp_transfrom_query, timestamp_server_response, timestamp_frame_decoded, timestamp_command_construct, timestamp_command_executed);
    });
    this->transport->set_on_transport_error([this]()
    {
        this->on_transport_error();
    });

    //this->transport->set_max_bitrate_send(this->send_bit_rate_limit);
    this->transport->set_max_send_queue_size(this->send_queue_size_limit * 1000);

    if (!this->transport->create(this->server_port))
    {
        return false;
    }

    return true;
}

bool RemoteHeadset::create_framebuffers()
{
    this->framebuffers.resize(this->frame_id_count);

    this->framebuffer_array = lava::make_image(this->get_format());
    this->framebuffer_array->set_layer_count(this->framebuffers.size());
    this->framebuffer_array->set_view_type(VK_IMAGE_VIEW_TYPE_2D_ARRAY);
    this->framebuffer_array->set_usage(VK_IMAGE_USAGE_SAMPLED_BIT | VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT | VK_IMAGE_USAGE_TRANSFER_SRC_BIT | VK_IMAGE_USAGE_TRANSFER_DST_BIT);

    if (!this->framebuffer_array->create(this->get_application()->get_device(), this->resolution))
    {
        return false;
    }

    for (uint32_t index = 0; index < this->framebuffers.size(); index++)
    {
        lava::image::ptr framebuffer = lava::make_image(this->get_format(), this->framebuffer_array->get());
        framebuffer->set_base_array_layer(index);
        framebuffer->set_view_type(VK_IMAGE_VIEW_TYPE_2D);

        if (!framebuffer->create(this->get_application()->get_device(), this->resolution))
        {
            return false;
        }

        this->framebuffers[index] = framebuffer;
    }

    return true;
}

void RemoteHeadset::destroy_framebuffers()
{
    for (lava::image::ptr& framebuffer : this->framebuffers)
    {
        framebuffer->destroy(true);
    }

    this->framebuffers.clear();

    if (this->framebuffer_array != nullptr)
    {
        this->framebuffer_array->destroy();
        this->framebuffer_array = nullptr;
    }
}

bool RemoteHeadset::create_encoders()
{
    this->encoders.resize(this->frame_id_count);

    for (uint32_t index = 0; index < this->encoders.size(); index++)
    {
        Encoder::Ptr encoder = make_encoder();

        lava::device_ptr device = this->get_application()->get_device();
        lava::renderer& renderer = this->get_application()->get_renderer();
        uint32_t frame_count = this->get_application()->get_frame_count();

        if (!encoder->create(device, renderer, this->transport->get_context().get_executor(), this->resolution, frame_count))
        {
            return false;
        }

        encoder->set_mode((EncoderMode)this->encoder_mode);
        encoder->set_key_rate(this->encoder_key_rate);
        encoder->set_bit_rate(this->encoder_bit_rate);
        encoder->set_frame_rate(this->encoder_frame_rate);
        encoder->set_quality_iframe(this->encoder_quality_iframe);
        encoder->set_quality_pframe(this->encoder_quality_pframe);

        this->encoders[index] = encoder;
    }

    return true;
}

void RemoteHeadset::destroy_encoders()
{
    for (uint32_t index = 0; index < this->encoders.size(); index++)
    {
        this->encoders[index]->destroy();
    }

    this->encoders.clear();
}

void RemoteHeadset::update_values()
{
    std::unique_lock<std::mutex> lock(this->transport_mutex);

    this->head_to_eye_matrices = this->transport_head_to_eye_matrices;
    this->projection_matrices = this->transport_projection_matrices;
    this->controller_states = this->transport_controller_states;
}

void RemoteHeadset::update_bitrates()
{
    this->statistic_send_bitrate->add_sample(this->transport->get_bitrate_send());
    this->statistic_receive_bitrate->add_sample(this->transport->get_bitrate_received());

    this->statistic_send_queue_size->add_sample(this->transport->get_send_queue_size() / 1000.0f);
    this->statistic_receive_queue_size->add_sample(this->transport->get_receive_queue_size() / 1000.0f);
}

void RemoteHeadset::update_stage(lava::delta delta_time)
{
    std::unique_lock<std::mutex> lock(this->transport_mutex);

    for (uint32_t index = 0; index < this->controller_buttons.size(); index++)
    {
        const std::array<PacketButtonState, BUTTON_ID_MAX_COUNT>& buttons = this->controller_buttons[index];
        glm::vec2 thumbstick = this->controller_thumbsticks[index];

        glm::mat4 controller_transform = this->controller_transforms[index];
        glm::vec3 controller_forward = controller_transform[2];
        glm::vec3 controller_sideward = controller_transform[0];

        float strength = this->movement_speed * delta_time;

        if (buttons[BUTTON_ID_TRIGGER] == BUTTON_STATE_PRESSED)
        {
            strength *= 2.0f;
        }

        stage_position += controller_forward * thumbstick.y * strength;
        stage_position -= controller_sideward * thumbstick.x * strength;
    }

    glm::mat4 stage_matrix = glm::translate(glm::mat4(1.0f), this->stage_position);
    this->view_matrix = this->head_matrix * stage_matrix;

    for (uint32_t index = 0; index < this->controller_matrices.size(); index++)
    {
        this->controller_matrices[index] = glm::inverse(stage_matrix) * this->controller_transforms[index];
    }

    lock.unlock();

    this->transport->send_stage_transform(stage_matrix);
}

void RemoteHeadset::on_setup(const glm::u32vec2& resolution)
{
    this->transport->send_setup_complete(this->strategy_id, this->frame_id_count, this->near_plane, this->far_plane);

    std::unique_lock<std::mutex> lock(this->transport_mutex);
    this->resolution = resolution;
    lock.unlock();
}

void RemoteHeadset::on_projection_change(const glm::mat4& left_eye_projection, const glm::mat4& right_eye_projection, const glm::mat4& left_head_to_eye, const glm::mat4& right_head_to_eye)
{
    std::unique_lock<std::mutex> lock(this->transport_mutex);
    this->transport_head_to_eye_matrices[EYE_LEFT] = left_head_to_eye;
    this->transport_head_to_eye_matrices[EYE_RIGHT] = right_head_to_eye;
    this->transport_projection_matrices[EYE_LEFT] = left_eye_projection;
    this->transport_projection_matrices[EYE_RIGHT] = right_eye_projection;
    lock.unlock();
}

void RemoteHeadset::on_head_transform(PacketTransformId transform_id, const glm::mat4& head_transform)
{
    std::unique_lock<std::mutex> lock(this->transport_mutex);
    this->transform_id = transform_id;
    this->head_matrix = head_transform;
    lock.unlock();
}

void RemoteHeadset::on_controller_transform(PacketControllerId controller_id, const glm::mat4& controller_transform)
{
    std::unique_lock<std::mutex> lock(this->transport_mutex);
    this->controller_transforms[controller_id] = controller_transform;
    lock.unlock();
}

void RemoteHeadset::on_controller_event(PacketControllerId controller_id, PacketControllerState controller_state)
{
    std::unique_lock<std::mutex> lock(this->transport_mutex);
    this->transport_controller_states[controller_id] = controller_state;
    lock.unlock();
}

void RemoteHeadset::on_controller_button(PacketControllerId controller_id, PacketButtonId button_id, PacketButtonState button_state)
{
    std::unique_lock<std::mutex> lock(this->transport_mutex);
    this->controller_buttons[controller_id][button_id] = button_state;
    lock.unlock();
}

void RemoteHeadset::on_controller_thumbstick(PacketControllerId controller_id, const glm::vec2& thumbstick)
{
    std::unique_lock<std::mutex> lock(this->transport_mutex);
    this->controller_thumbsticks[controller_id] = thumbstick;
    lock.unlock();
}

void RemoteHeadset::on_latency(PacketFrameNumber frame_number, PacketFrameId frame_id, PacketTransformId transform_id, double timestamp_transform_query, double timestamp_server_response, double timestamp_frame_decoded, double timestamp_command_construct, double timestamp_command_executed)
{
    LatencyCapture capture;
    capture.frame_id = frame_id;
    capture.frame_number = frame_number;
    capture.transform_id = transform_id;
    capture.timestamp_transform_query = timestamp_transform_query;
    capture.timestamp_server_response = timestamp_server_response;
    capture.timestamp_frame_decoded = timestamp_frame_decoded;
    capture.timestamp_command_construct = timestamp_command_construct;
    capture.timestamp_command_executed = timestamp_command_executed;

    std::unique_lock<std::mutex> lock(this->transport_mutex);
    this->statistic_latency->submit_capture(capture);
    lock.unlock();
}

void RemoteHeadset::on_transport_error()
{
    lava::log()->error("Remote Headset: Transport error detected!");
}

bool RemoteHeadset::convert_strategy(StereoStrategyType strategy_type, PacketStereoStrategyId& strategy_id)
{
    switch (strategy_type)
    {
    case STEREO_STRATEGY_TYPE_NATIVE_FORWARD:
        strategy_id = STEREO_STRATEGY_ID_NATIVE_STEREO;
        break;
    case STEREO_STRATEGY_TYPE_NATIVE_DEFERRED:
        strategy_id = STEREO_STRATEGY_ID_NATIVE_STEREO;
        break;
    case STEREO_STRATEGY_TYPE_MULTI_VIEW:
        strategy_id = STEREO_STRATEGY_ID_NATIVE_STEREO;
        break;
    default:
        lava::log()->error("Remote Headset: Unkown strategy type!");
        return false;
    }

    return true;
}